__author__ = 'Qi'
# Created by on 11/25/22.
import torch
import torch.nn as nn

class ABLoss(nn.Module):
    '''
    This is the attention biased loss for handling data imbalance problem.
    When combining with SGD optimizer, its the absgd.
    When combining with Adam optimizer, its the ABAdam.
    We implement the multi stage lambda that: applying the ABSGD after learning a representation at certain epoch.
    '''
    def __init__(self,  mylambda, milestone = 0, criterion = nn.CrossEntropyLoss(reduction='none'), abgamma = 0.9, abalpha =1):
        '''
        :param mylambda: regularizers that we want to set
        :param milestone (default 0): decayed epoch
        :param criterion (default multi-class cross entropy): loss criterion
        :param abgamma (default 0.9) : moving average parameter
        :param abalpha (default 1) : control the stability of moving average parameter, (0, 1]
        '''
        super(ABLoss, self).__init__()
        self.u = 0
        self.abgamma = abgamma
        self.abalpha = abalpha
        self.criterion = criterion
        self.milestone = milestone
        self.ablambda = None
        self.mylambda = mylambda
        self.step = 0

    def forward(self, output, target):
        loss = self.criterion(output, target)
        if self.ablambda is None:
            # reduces to CE
            p = 1/len(loss)
        else:
            # exponential loss
            expLoss = torch.exp(loss / self.ablambda)
            # record moving average history
            self.u = (1 - self.abgamma) * self.u + self.abgamma * (self.abalpha * torch.mean(expLoss))
            # robust weights
            drop = expLoss/(self.u * len(loss))
            drop.detach_()
            p = drop

        abloss = torch.sum(p * loss)
        return abloss

    def updateLambda(self):
        if self.step >= self.milestone - 1:
            self.ablambda = self.mylambda
        self.step += 1







__author__ = 'Qi'
# Created by on 11/25/22.

import torch
import torchvision
import torchvision.transforms as transforms
import numpy as np
import os
import torch.utils.data as data
from PIL import Image
import json
from torch.utils.data import Dataset
from collections import Counter


class IMBALANCECIFAR10(torchvision.datasets.CIFAR10):
    cls_num = 10

    def __init__(self, root, imb_type='exp', imb_factor=0.01, rand_number=0, train=True,
                 transform=None, target_transform=None,
                 download=True):
        super(IMBALANCECIFAR10, self).__init__(root, train, transform, target_transform, download)
        np.random.seed(rand_number)
        img_num_list = self.get_img_num_per_cls(self.cls_num, imb_type, imb_factor)
        self.gen_imbalanced_data(img_num_list)

    def get_img_num_per_cls(self, cls_num, imb_type, imb_factor):
        img_max = len(self.data) / cls_num

        img_num_per_cls = []
        if imb_type == 'exp':
            for cls_idx in range(cls_num):
                num = img_max * (imb_factor ** (cls_idx / (cls_num - 1.0)))
                img_num_per_cls.append(int(num))
        elif imb_type == 'step':
            for cls_idx in range(cls_num // 2):
                img_num_per_cls.append(int(img_max))
            for cls_idx in range(cls_num // 2):
                img_num_per_cls.append(int(img_max * imb_factor))
        else:
            img_num_per_cls.extend([int(img_max)] * cls_num)

        return img_num_per_cls

    def gen_imbalanced_data(self, img_num_per_cls):
        new_data = []
        new_targets = []
        targets_np = np.array(self.targets, dtype=np.int64)
        classes = np.unique(targets_np)
        # np.random.shuffle(classes)
        self.num_per_cls_dict = dict()
        for the_class, the_img_num in zip(classes, img_num_per_cls):
            self.num_per_cls_dict[the_class] = the_img_num
            idx = np.where(targets_np == the_class)[0]
            np.random.shuffle(idx)
            selec_idx = idx[:the_img_num]
            new_data.append(self.data[selec_idx, ...])
            new_targets.extend([the_class, ] * the_img_num)
        new_data = np.vstack(new_data)
        self.data = new_data
        self.targets = new_targets

    def get_cls_num_list(self):
        cls_num_list = []
        for i in range(self.cls_num):
            cls_num_list.append(self.num_per_cls_dict[i])
        return cls_num_list

        # def _check_integrity(self):
        #     return True


class IMBALANCECIFAR100(IMBALANCECIFAR10):
    """`CIFAR100 <https://www.cs.toronto.edu/~kriz/cifar.html>`_ Dataset.
    This is a subclass of the `CIFAR10` Dataset.
    """
    base_folder = 'cifar-100-python'
    url = "https://www.cs.toronto.edu/~kriz/cifar-100-python.tar.gz"
    filename = "cifar-100-python.tar.gz"
    tgz_md5 = 'eb9058c3a382ffc7106e4002c42a8d85'
    train_list = [
        ['train', '16019d7e3df5f24257cddd939b257f8d'],
    ]

    test_list = [
        ['test', 'f0ef6b0ae62326f3e7ffdfab6717acfc'],
    ]
    meta = {
        'filename': 'meta',
        'key': 'fine_label_names',
        'md5': '7973b15100ade9c7d40fb424638fde48',
    }
    cls_num = 100


    # def _check_integrity(self):
    #     return True



__author__ = 'Qi'
# Created by on 11/25/22.
import torch

def ABSGD(model_param, lr, weight_decay = 0.0):
    return torch.optim.SGD(model_param, lr = lr,  momentum=0.9, weight_decay=weight_decay)


def ABAdam(model_param, lr, weight_decay= 0.0):
    return torch.optim.Adam(model_param, lr = lr, weight_decay=weight_decay)
